# coding=utf-8
"""Скрипт прогрева кеша тайлов."""
import math
import requests

from datetime import datetime
from multiprocessing.dummy import Pool as ThreadPool
from urllib.parse import (
    parse_qs,
    urlparse,
    ParseResult,
)

from omlibs.v1 import (
    GLog,
    OMSApi,
    Checks
)

from omlibs.v1.env import (
    LOG_SCRIPT_CODE,
)

Z_RANGE_MIN = 2
Z_RANGE_MAX = 19
BBOX_X_MIN = -180
BBOX_X_MAX = 180
BBOX_Y_MIN = -90
BBOX_Y_MAX = 90

MIN_POOL_SIZE = 1
MAX_POOL_SIZE = 8

# multiprocessing.dummy.Pool сжирает память на миллионых очередях, попробуем ограничить
# большой размер пула может породить ситуацию, когда в его конце окажутся тайлы с новым токеном,
# и он просрочится раньше, чем до него дойдет очередь
# todo произвести замеры на разных размерах очередей
URL_POOL_LIMIT = 1000

# частота обновления инфорамции по прогреву, каждые N секунд
TASK_UPDATE_FREQUENCY = 30

# шаблон статистики выполнения прогрева
TASK_STAT_PROCESS_TPL = "Осталось %(days)sд %(hours)sч %(mins)sм %(secs)sс, скорость: %(speed)s тайлов/сек"
TASK_STAT_FINISH_TPL = "Выполнено за %(days)sд %(hours)sч %(mins)sм %(secs)sс, скорость: %(speed)s тайлов/сек"

# месторасположение списка задач прогрева кеша
task_map_code = "system"
task_layer_code = "cache_burner"

oms_api = OMSApi()


def get_possible_task(lock_id: str) -> dict:
    """Получение детальной инфомации по допустимой задаче прогрева.

    :param lock_id: идентификатор блокировки скрипта
    """
    task_id = None
    task = {}

    all_tasks = oms_api.get_layer_objects(task_map_code, task_layer_code, "lock_id,done")
    for p_task in all_tasks:
        p_task_lock_id = p_task["fields"]["lock_id"]
        # подходит задача не завершенная и не заблокированная другим скриптом
        if not p_task["fields"]["done"] and (not p_task_lock_id or p_task_lock_id == lock_id):
            task_id = p_task["id"]
            break

    if task_id:
        task = oms_api.get_layer_object(task_map_code, task_layer_code, task_id)

    return task


def parse_task_params(tile_url_example: str = None, bbox_burn: str = None, z_range: str = None, login_list: str = None,
                      metatile: str = None, **kwargs) -> (str, list, list, list, list):
    """Подготовка параметров задачи.

    :param tile_url_example пример URL тайла
    :param bbox_burn:       bbox области прогрева
    :param z_range:         диапазон масштабов прогрева
    :param login_list:      логины пользователй для прогрева
    :param metatile:        размер метатайла, учитываемый при прогреве
    """
    # пример: "http://lm.orbismap.com/tms/oms2_dev/m7838/?layers=871&x=77&y=87&z=7&base=0&q=1
    # &token=5e5d5489291573b47cc5ada10c49a7fb32d1ae82"
    # в результате: "http://lm.orbismap.com/tms/oms2_dev/m7838/?layers=871&x={x}&y={y}&z={z}&base=0&q=1"
    url_obj = urlparse(tile_url_example)
    # значения параметров не бывает множественными, так что выберем только первые из списков
    url_query = {k: v[0] for k, v in parse_qs(url_obj.query).items()}
    # токен индивидуально генерируется по переданным пользователям в процессе формирования итогового URL тайла
    url_query.pop("token", None)
    # координаты будут подставляться в процессе формирования итогого URL тайла
    for k in ["x", "y", "z"]:
        if k in url_query:
            url_query[k] = "{" + k + "}"

    url_base_args = {
        "scheme": url_obj.scheme,
        "netloc": url_obj.netloc,
        "path": url_obj.path,
        "params": url_obj.params,
        "query": None,
        "fragment": url_obj.fragment,
    }
    url_base = ParseResult(**url_base_args).geturl()
    # формирование query_string без encoding, т.к. нужно сохранить подстановки {x|y|z}
    res_tile_url_template = url_base + "?" + "&".join([k + "=" + v for k, v in url_query.items()])

    # остальные параметры только разбираем, проверка будет в основном методе
    # пример: "36.7850, 55.3900, 38.4933, 56.1410"
    # в результате: [36.7850, 55.3900, 38.4933, 56.1410]
    res_bbox = [float(i.strip()) for i in bbox_burn.split(",")] if bbox_burn else []
    # пример: "6-9"
    # в результате: [6, 9]
    res_z_range = [int(i.strip()) for i in z_range.split("-")] if z_range else []
    # пример: ",lm,lm2,lm2"
    # в результате: ["", "lm", "lm2"]
    res_login_list = list(set([i.strip() for i in login_list.split(",")] if login_list else [""]))
    # пример: "3,2"
    # в результате: [3, 2]
    res_metatile = [int(i.strip()) for i in metatile.split(",")] if metatile else []

    GLog.debug('tile_url_template: %s', res_tile_url_template)
    GLog.debug('bbox: %s', res_bbox)
    GLog.debug('z_range: %s', res_z_range)
    GLog.debug('login_list: %s', res_login_list)
    GLog.debug('metatile: %s', res_metatile)

    return res_tile_url_template, res_bbox, res_z_range, res_login_list, res_metatile


def check_bbox(bbox: list) -> bool:
    """Проверка bbox на корректность значения.

    :param bbox: Bbox
    """
    # todo корректнее проверить для 4326
    # bbox в формате  [mixX, minY, maxX, maxY]
    if len(bbox) != 4 or (bbox[0] >= bbox[2] or bbox[1] >= bbox[3]) or \
            (bbox[0] < BBOX_X_MIN or bbox[1] < BBOX_Y_MIN or bbox[2] >= BBOX_X_MAX or bbox[3] > BBOX_Y_MAX):
        GLog.error("Некорректный bbox: %s" % str(bbox))
        return False

    return True


def check_z_range(z_range: list) -> bool:
    """Проверка диапазона масштабов на корректность значения.

    :param z_range: Диапазон масштабов
    """
    # z_range в формате  [mixZ, maxZ], используютс от 2 до 19 в проектах
    if len(z_range) != 2 or (z_range[0] > z_range[1]) or (z_range[0] < Z_RANGE_MIN or z_range[1] > Z_RANGE_MAX):
        GLog.error("Некорректный z_range: %s" % str(z_range))
        return False
    return True


def check_metatile(metatile: list) -> bool:
    """Проверка размера метатайла на корректность значения.

    :param metatile: Размер метатайла
    """
    # metatile в формате  [width, height]
    if len(metatile) != 2 or metatile[0] < 1 or metatile[1] < 1:
        GLog.error("Некорректный metatile: %s" % str(metatile))
        return False
    return True


def deg2num(lon_deg: float, lat_deg: float, zoom: int) -> (int, int):
    """Перевод долготы и широты в x,y координаты тайла.
    http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Lon..2Flat._to_tile_numbers_2

    :param lon_deg: Долгота в градусах
    :param lat_deg: Широта в градусах
    :param zoom: Масштаб
    """
    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    # корректировка Y координаты, инвертирована
    ytile = int(n - 1) - int((1.0 - math.log(math.tan(lat_rad) + (1 / math.cos(lat_rad))) / math.pi) / 2.0 * n)
    return xtile, ytile


def generate_urls(tile_url_template: str, login_list: list, z_range: list, bbox: list, metatile: list):
    """Формирование url тайлов через генератор.

    :param tile_url_template    шаблон URL тайла
    :param login_list:          логины пользователй для прогрева
    :param z_range:             диапазон масштабов прогрева
    :param bbox:                bbox области прогрева
    :param metatile:            размер метатайла, учитываемый при прогреве
    """
    # попользовательно, помасштабно, подолготно и поширотно формируем url тайлов
    # (перепрыгивания значения по x,y с учетом заданного метатайла)
    for login in login_list:
        token = oms_api.get_token(login) if login else ""
        for z in range(z_range[0], z_range[1] + 1):
            x_start, y_start = deg2num(bbox[0], bbox[1], z)
            x_end, y_end = deg2num(bbox[2], bbox[3], z)
            x_current = x_start
            # увеличить диапазон прогреваемых тайлов на случай попадания в нижний правый угол метатайла
            # если не увеличить диапазон, то случится пропуск тайлов с x_end и y_end
            x_end += metatile[0]
            y_end += metatile[1]
            while x_current < x_end:
                y_current = y_start
                while y_current < y_end:
                    tile_url = tile_url_template.format(x=x_current, y=y_current, z=z)
                    if token:
                        tile_url += "&token=" + token
                    yield tile_url
                    y_current += metatile[1]
                x_current += metatile[0]


def process_url(tile_url: str, burn_stat: dict, task: dict):
    """Запрос тайла со сбором статистики.

    Статистка прогрева  содержит изменяемую и опорную части:
        {
            'tiles_processed': 0, # int, изменяемое, для подсчета обработанных тайлов
            'metatile': [3,2], # list[int, int]
            'task_update_counter': 0, # int
            'started_at': ..., # datetime.datetime
        }
    :param tile_url:    URL тайлa для прогрева
    :param burn_stat:   Статистика прогрева
    :param task:        Обрабатываемая задача
    """
    requests.head(tile_url)
    burn_stat["tiles_processed"] += burn_stat["metatile"][0] * burn_stat["metatile"][1]
    # т.к. время обработки тайла недетерминировано, то момент обновления статистики по задаче
    # не получить через остаток деления прошедшего времени на TASK_UPDATE_FREQUENCY, нулевой остаток не гарантирован
    # поэтому отслеживаем изменение целочисленного частного
    # ввиду конкурентности выполнения данного метода через pool, статистика может выдавать несколько неточные данные
    # время обработки текущей задачи
    exec_time = datetime.now() - burn_stat["started_at"]
    exec_seconds = exec_time.days * 24 * 3600 + exec_time.seconds
    if exec_seconds // TASK_UPDATE_FREQUENCY > burn_stat["task_update_counter"]:
        burn_stat["task_update_counter"] = exec_seconds // TASK_UPDATE_FREQUENCY
        GLog.info("Тайлов обработано: %s" % burn_stat["tiles_processed"])
        task["fields"]["tiles_processed"] = burn_stat["tiles_processed"]
        task["fields"]["updated_at"] = str(datetime.now())
        # предсказать время и скорость работы можно пока тайлов обработано меньше,
        # чем всего (при использовании метатайла тайлы обрабатываются с избытком)
        if task["fields"]["tiles_total"] > task["fields"]["tiles_processed"]:
            # средняя скорость обработки со старта задачи
            # todo кешированне тайлы портят статистику
            tiles_per_second = task["fields"]["tiles_processed"] // exec_seconds
            # прогноз оставшегося времени
            # время некорректное при обработке медленнее 1 тайла в сек
            estimate_seconds = (task["fields"]["tiles_total"] - task["fields"]["tiles_processed"]) // (
                tiles_per_second if tiles_per_second else 1)
            task["fields"]["stat"] = build_task_stat_info(TASK_STAT_PROCESS_TPL, tiles_per_second, estimate_seconds)
        oms_api.edit_layer_object(task_map_code, task_layer_code, task["id"], task)


def build_task_stat_info(tpl: str, tiles_per_second: int, stat_time: int) -> str:
    """Формирование отчета по статистике выполнения скрипта.

    :param tpl:                 Шаблон отчета
    :param tiles_per_second:    Скорость выполнения, тайлов в секунду
    :param stat_time: Время     для отчета
    """
    return tpl % {
        "speed": tiles_per_second,
        "days": stat_time // (24 * 3600),
        "hours": (stat_time // 3600) % 24,
        "mins": (stat_time // 60) % 60,
        "secs": stat_time % 60,
    }


def process_url_pool(url_pool: list, pool_size: int, burn_stat: dict, task: dict):
    """Параллельная обработка запросов тайлов.

    :param url_pool:    Список URL тайлов для параллельного прогрева
    :param pool_size:   Кол-во одновремнных запросов
    :param burn_stat:   Статистика прогрева
    :param task:        Обрабатываемая задача
    """
    if not url_pool:
        return

    pool = ThreadPool(pool_size)
    pool.map(lambda url: process_url(url, burn_stat, task), url_pool)
    pool.close()
    pool.join()


def get_pool_size(task: dict) -> int:
    """Получить параметр параллельности прогрева кэша

    :param task: задача на прогрев
    :return: кол-во потоков
    """
    concurrent = task['fields'].get('concurrent', 0)
    if not concurrent:
        GLog.debug('url_pool_size: %d' % MIN_POOL_SIZE)
        return MIN_POOL_SIZE

    _arg = 'кол-во одновременных запросов'
    url_pool_size = Checks.check_int(concurrent, _arg)
    _msg = "Некорректное кол-во одновременных запросов (%s), допустимо от %s до %s"
    if url_pool_size < MIN_POOL_SIZE:
        GLog.info(_msg % (url_pool_size, MIN_POOL_SIZE, MAX_POOL_SIZE))
        url_pool_size = MIN_POOL_SIZE
    elif url_pool_size > MAX_POOL_SIZE:
        GLog.info(_msg % (url_pool_size, MIN_POOL_SIZE, MAX_POOL_SIZE))
        url_pool_size = MAX_POOL_SIZE

    GLog.debug('url_pool_size: %d' % url_pool_size)
    return url_pool_size


def calc_total_tiles_count(z_range: list, bbox: list, login_list: list) -> int:
    """Получить суммарное кол-во тайлов, которое необходимо запросить

    :param list z_range: диапазон масштабов
    :param list bbox: габарит карты для прогрева
    :param list login_list: список логинов для прогрева
    :return: суммарное кол-во тайлов
    """
    tiles_total = 0
    for z in range(z_range[0], z_range[1] + 1):
        x_min, y_min = deg2num(bbox[0], bbox[1], z)
        x_max, y_max = deg2num(bbox[2], bbox[3], z)
        tiles_total += (x_max - x_min + 1) * (y_max - y_min + 1)

    tiles_total *= len(login_list)
    GLog.debug('tiles_total: %d' % tiles_total)
    return tiles_total


def run():
    """."""
    # блокировка задачи по коду скрипта, а не через переменные скрипта
    # т.к. вместе с дублированием скрипта в новый скопировалась бы и его блокировка
    lock_id = LOG_SCRIPT_CODE
    GLog.info('Lock ID: %s' % lock_id)

    # выбрать записи из таблицы задач прогрева
    # выбрать первую не заблокированную (кроме собственного lock_id) и не завершенную задачу
    # выбрать по ней детальную информацию и заблокировать по lock_id
    # выполнять подряд все доступные задачи
    while True:
        task = get_possible_task(lock_id)
        if not task:
            GLog.info("Отсутствуют задачи для выполнения")
            return 0

        # todo html символы
        if "tile_url_example" in task["fields"]:
            task["fields"]["tile_url_example"] = task["fields"]["tile_url_example"].replace("&amp;", "&")
        GLog.info(task)
        if not task["fields"]["lock_id"]:
            task["fields"]["lock_id"] = lock_id
            oms_api.edit_layer_object(task_map_code, task_layer_code, task["id"], task)
            GLog.info("Добавлена блокировка к задаче")

        # обработать параметры задачи
        tile_url_template, bbox, z_range, login_list, metatile = parse_task_params(**task["fields"])

        # использовать пул запросов к тайлам при необходимости
        url_pool_size = get_pool_size(task)

        # проверить параметры задачи
        check = check_bbox(bbox) and check_z_range(z_range) and check_metatile(metatile)
        if not check:
            return -1

        # просчитать количество тайлов на обработку и обновить запись задачи
        # попутно обнулив кол-во обработанных
        tiles_total = calc_total_tiles_count(z_range, bbox, login_list)

        task["fields"]["tiles_total"] = tiles_total
        task["fields"]["tiles_processed"] = 0
        task["fields"]["updated_at"] = str(datetime.now())
        task["fields"]["stat"] = ""
        oms_api.edit_layer_object(task_map_code, task_layer_code, task["id"], task)
        GLog.info("Рассчитано количество тайлов: %s" % tiles_total)

        burn_stat = {
            "tiles_processed": 0,
            "task_update_counter": 0,
            "metatile": metatile,
            "started_at": datetime.now(),
        }
        # прогрев через пул запросов тайлов
        # pool = ThreadPool(url_pool_size)
        # pool.map(lambda url: process_url(url, burn_stat, task),
        #          generate_urls(tile_url_template, login_list, z_range, bbox, metatile))
        # pool.close()
        # pool.join()

        # разбить весь список на относительно небольшие, т.к. на миллионах Pool кушает память
        url_pool = []
        for url in generate_urls(tile_url_template, login_list, z_range, bbox, metatile):
            url_pool.append(url)
            if len(url_pool) == URL_POOL_LIMIT:
                process_url_pool(url_pool, url_pool_size, burn_stat, task)
                url_pool = []

        # возможные необработанные остатки
        process_url_pool(url_pool, url_pool_size, burn_stat, task)

        GLog.info("Тайлов обработано, итого: %s" % burn_stat["tiles_processed"])

        # как только всё завершили - снимаем блокировку и помечаем задачу как завершенную, обновялем статистику
        task["fields"]["tiles_processed"] = burn_stat["tiles_processed"]
        task["fields"]["updated_at"] = str(datetime.now())
        task["fields"]["lock_id"] = ""
        task["fields"]["done"] = True
        # время обработки текущей задачи
        exec_time = datetime.now() - burn_stat["started_at"]
        exec_seconds = exec_time.days * 24 * 3600 + exec_time.seconds
        if exec_seconds:
            # средняя скорость обработки со старта задачи
            # todo кешированне тайлы портят статистику
            tiles_per_second = task["fields"]["tiles_processed"] // exec_seconds
            task["fields"]["stat"] = build_task_stat_info(TASK_STAT_FINISH_TPL, tiles_per_second, exec_seconds)
        oms_api.edit_layer_object(task_map_code, task_layer_code, task["id"], task)
        GLog.info("Снята блокировка, задача выполнена")


def main():
    """Точка входа.

    Выполнение скрипта интерпретируется в вызов данной функции
    """
    try:
        return run()
    except Exception as e:
        GLog.exception('Ошибка: ', e)
        return -1
